import os
import re
import math
import itertools
from functools import partial
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

from rosemary import plt_scaled_colobar_ax

from evaluate import get_eval_results, jpt_default_display
from calibration import plt_reliability_diagram_subplot, get_reliability_diagram_subplot_kwargs, calibrate_ot_on_task_result



def df_to_latex(
        df,
        column_format=None,
        formatter=None,
    ):

    if column_format is None:
        column_format = 'l'*len(df.columns)
    s = (df.style
        .hide(axis='index') 
        .format(formatter, na_rep='')
        .to_latex(
            column_format=column_format,
            sparse_index=True,
            multirow_align='c',
            sparse_columns=True,
            multicol_align='c',
            position_float='centering',
            hrules=True,
            position='t',
        ))
    return s


def df_to_latex_with_colors(
        df,
        cmap='RdYlGn_r',
        subset=None,
        low=0.2,
        high=0.8,
        alpha=.5,
        higher_better=None,
        column_format=None,
    ):
    # """Remember to add the following to color the table cell 
    #         \usepackage{colortbl}"""
    from matplotlib.colors import to_rgb
    def rgb_to_latex(rgb):
        return f"{rgb[0]},{rgb[1]},{rgb[2]}"

    if higher_better is None:
        higher_better = {col: True for col in subset}
    elif isinstance(higher_better, bool):
        higher_better = {col: higher_better for col in subset}
    
    # Create a copy of the DataFrame to avoid modifying the original
    styled_df = df.copy()
    
    # Create the colormap
    if isinstance(cmap, str):
        cmap = plt.get_cmap(cmap)
    
    # Function to create cell content with color
    def color_cell(val, color):
        return f"\\cellcolor[rgb]{{{color}}}{val}"

    # Function to apply alpha to RGB color
    def apply_alpha(rgb, alpha):
        return tuple(alpha * c + (1 - alpha) for c in rgb)
    
    # Apply color scaling to each column
    for col in subset:
        series = styled_df[col]
        vmin, vmax = series.quantile(low), series.quantile(high)
        if higher_better[col]:
            normalized = (series - vmin) / (vmax - vmin)
        else:
            normalized = 1 - (series - vmin) / (vmax - vmin)
        normalized = normalized.clip(0, 1)  # Clip values to [0, 1] range
        colors = normalized.map(lambda x: rgb_to_latex(apply_alpha(to_rgb(cmap(x)), alpha)))
        styled_df[col] = styled_df[col].astype(str).combine(colors, color_cell)
    
    # Set up the Styler object
    styler = styled_df.style.hide(axis='index')
    
    # Set default column format if not provided
    if column_format is None:
        column_format = 'l' * len(df.columns)
    
    # Convert to LaTeX
    latex_table = styler.to_latex(
        column_format=column_format,
        sparse_index=True,
        multirow_align='c',
        sparse_columns=True,
        multicol_align='c',
        position_float='centering',
        hrules=True,
        position='t',
    )
    
    return latex_code


def replace_specific_occurrences(s, substr, occurrences_to_replace):
    """0-index. """
    occurrences_to_replace = [x+1 for x in occurrences_to_replace]
    
    # Find all occurrences of the substring
    matches = list(re.finditer(re.escape(substr), s))
    
    # Sort occurrences to replace and reverse to process them from the end of the string
    occurrences_to_replace = sorted(occurrences_to_replace, reverse=True)
    
    # Convert the string into a list to allow easy modification
    s_list = list(s)
    
    for occurrence in occurrences_to_replace:
        if occurrence <= len(matches):
            # Get the start and end index of the substring to replace
            start, end = matches[occurrence - 1].span()
            # Replace the substring with empty string
            s_list[start:end] = ''
    
    # Join the modified list back into a string
    return ''.join(s_list)


def to_latex_bold_formatter(x, value, num_decimals=2):
    """Format a number in bold when (almost) identical to a given value.
            where value is the value to compare x with. 
            Returns the formatted string. 
            
        Consider values equal, when rounded results are equal
            otherwise, it may look surprising in the table where they seem identical

        Need to define macro textBF according to:
            https://tex.stackexchange.com/questions/23678/textbftext-without-increasing-the-length-of-the-text
    """
    if round(x, num_decimals) == round(value, num_decimals):
        return f"\\textbf{{{x:.{num_decimals}f}}}"
        # return f"{{\\bfseries{x:.{num_decimals}f}}}"
    else:
        return f"{x:.{num_decimals}f}"


def get_default_bold_formatter(
        df,
        cols_take_max=[],
        cols_take_min=[],
    ):
    formatter = {}
    for col in df.select_dtypes(include=[np.number]):
        num_decimals = 2
        metric_name = col[1]
        if metric_name in cols_take_max:
            value = df[col].max()
        elif metric_name in cols_take_min:
            value = df[col].min()
        else:
            raise ValueError(f'Not sure to take min/max to bold for column: {col}')
        formatter[col] = partial(to_latex_bold_formatter, value=value, num_decimals=num_decimals)
    return formatter


def get_string_between(s, a, b):
    try:
        start = s.find(a) + len(a)  # Find the starting index after 'a'
        end = s.find(b, start)  # Find the ending index of 'b' after 'a'
        
        # Check if 'a' and 'b' were found correctly and in order
        if start > len(a) - 1 and end != -1:
            return s[start:end].strip()  # Return the substring between 'a' and 'b'
        else:
            return ""  # Return empty string if 'a' or 'b' not found in correct positions
    except ValueError:
        return ""  # Return empty string if any error occurs


def plt_source_of_dcp_dist_prompt_gpt4o_vary_K(assets_path='./'):
    """Effect of the fixed certainty phrase set size (K) on LLM calibration performance.
    
        Specifically, ECE vs. #phrases plot. for the two tasks {truthfulqa, sciq} for {gpt-3.5-turbo, gpt-4o-mini, gpt-4o}
    """

    cols = ['Model', 'Task', r'$(u_1,\cdots,u_K)$ Source']
    cols += ['Correctness', 'ECE', 'BS']

    exp_dirs = [
        'results/gpt-3.5-turbo-0125',
        'results/gpt-4o-mini',
        'results/gpt-4o-2024-08-06',
    ]

    run_path_filter_fn = lambda x: ('p=verbconf_' in x) and ('src=prompt:gpt-4o:size=' in x)

    df = get_eval_results(
        exp_dirs=exp_dirs,
        cols=cols,
        run_path_filter_fn=run_path_filter_fn,
        group_metrics_by_task=True,
        sort_df=None,
    )['df']

    dcp_source_model_name = 'gpt-4o'
    Ks = [4, 8, 12, 16, 20]
    dcp_source = [f"prompt {dcp_source_model_name} " + f'($K={K}$)' for K in Ks]


    task_name_list = ['SciQ', 'TruthfulQA']
    model_name_list = ['gpt-3.5-turbo', 'gpt-4o-mini', 'gpt-4o']
    metric_name_list = ['ECE', 'BS']

    plt_params = {
        'font.size': 35,
        'font.family': 'Times New Roman',
        # "legend.frameon": False,
        "axes.titlepad": 10,
        'xtick.labelsize': 15,
        'ytick.labelsize': 15,
        'axes.prop_cycle': plt.cycler(color=plt.cm.tab20c.colors),
        'lines.linewidth': 3,
        'lines.markersize': 10,
    }

    with plt.rc_context(plt_params):

        nrows = len(metric_name_list)
        ncols = len(task_name_list)
        fig, axs = plt.subplots(nrows, ncols, figsize=(4*ncols, 4*nrows), sharey=True)

        for axi in range(nrows):
            for axj in range(ncols):
                ax = axs.reshape(nrows, ncols)[axi, axj]
                metric_name = metric_name_list[axi]
                task_name = task_name_list[axj]
            
                xs = Ks
                for model_name in model_name_list:    
                    dft = df[df['Model']==model_name]
                    dft.loc[:, 'rank'] = pd.Categorical(dft[r'$(u_1,\cdots,u_K)$ Source'], categories=dcp_source, ordered=True)
                    dft = dft.sort_values('rank')
                    ys = dft[(task_name, metric_name)].tolist()
                    ax.plot(xs, ys, 'o-', label=model_name)
        
                ax.set_xticks(Ks)
                ax.set_xlim((0, 21))
                ax.yaxis.set_major_locator(MaxNLocator(nbins=3, prune='both'))
                ax.grid()

        for axi in range(nrows):
            axs.reshape(nrows, ncols)[axi, 0].set_ylabel(metric_name_list[axi])
        for axj in range(ncols):
            axs.reshape(nrows, ncols)[nrows-1, axj].set_xlabel('K')
            axs.reshape(nrows, ncols)[0, axj].set_title(task_name_list[axj])

        
        # last ax's handle.
        handles, labels = ax.get_legend_handles_labels()
        fig.legend(handles, labels, loc='lower center', ncol=3, bbox_to_anchor=(0.5, -0.02), fontsize=18)
        fig.tight_layout()

    fig_path = os.path.join(assets_path, f'source_of_dcp_dist_prompt_gpt4o_vary_K.png')
    fig.savefig(fig_path, dpi=300, bbox_inches='tight')
    print(f'Saved to fig_path: {fig_path}')



def table_calibration_performance_vary_dcp_source():
    """ Calibration performance of LLMs prompted to verbalize confidence, where the certainty distributions $\{u_k\}_k$ are derived from \textit{survey}, generated \textit{on-the-fly} for each example, or chosen from a \textit{fixed} set ($K=12$) pre-generated with GPT-4o. Fixing certainty distributions significantly improve calibration.

    ```
    from paper_figures import table_calibration_performance_vary_dcp_source
    s = table_calibration_performance_vary_dcp_source()
    print(s)
    ```
    """

    cols = ['Model', 'Task', r'$(u_1,\cdots,u_K)$ Source']
    cols += ['Correctness', 'ECE', 'BS']

    exp_dirs = [
        'results/gpt-3.5-turbo-0125',
        'results/gpt-4o-mini',
        'results/gpt-4o-2024-08-06',
        'results/claude-3-haiku-20240307',
        'results/claude-3-5-sonnet-20240620',
        'results/gemini-1.5-flash-002',
        'results/gemini-1.5-pro-002',
    ]

    run_path_filter_fn = lambda x: \
        ('p=verbconf_src=prompt:gpt-4o:size=12' in x) or \
        ('p=verbconf_src=social:media:poll' in x) or \
        ('p=verbconfv3_src=unspecified:beta' in x)

    df = get_eval_results(
        exp_dirs=exp_dirs,
        cols=cols,
        run_path_filter_fn=run_path_filter_fn,
        group_metrics_by_task=True,
        sort_df=None,
    )['df']


    ordering = [
        "survey",
        "on-the-fly",
        "prompt gpt-4o ($K=12$)",
    ]

    model_name_list = [
        # 'gpt-3.5-turbo',
        # 'gpt-4o-mini',
        'gpt-4o',
        # 'claude-3-haiku',
        'claude-3.5-sonnet',
        # 'gemini-1.5-flash',
        'gemini-1.5-pro',
    ]

    L = []
    for model_name in model_name_list:

        dft = df[df['Model']==model_name].copy()
        rank = pd.Categorical(dft[r'$(u_1,\cdots,u_K)$ Source'], categories=ordering, ordered=True)
        has_nan = any([math.isnan(x) if isinstance(x, float) else False for x in rank.tolist()])
        if has_nan:
            raise ValueError(f'ranking has nan. rank: {rank}\n ordering: {ordering}')
        dft.loc[:, 'rank'] = rank
        dft = dft.sort_values('rank', ascending=True)
        dft = dft.drop(columns=['rank'])
        dft.reset_index(drop=True)
        
        ## bold cell   
        formatter = get_default_bold_formatter(
            dft,
            cols_take_max=['Acc', 'Correctness'],
            cols_take_min=['ECE', 'BS'],
        )
        s = df_to_latex(dft, 
                        column_format=None,
                        formatter=formatter,
                    )
        ## color cell
        from matplotlib.colors import to_rgb, LinearSegmentedColormap
        def create_custom_blue_cmap():
            return LinearSegmentedColormap.from_list("custom_blue", ['#FFFFFF', '#E6F3FF', '#99D6FF', '#4DB8FF'])
        cmap = create_custom_blue_cmap()
        higher_better = [x for x in list(dft.columns) if x[1]]
        higher_better = dict({x: (True if x[1] in ['Correctness'] else False) for x in higher_better })
        s = df_to_latex_with_colors(dft.round(2), cmap=cmap, alpha=.75, low=0, high=.7,
                                higher_better=higher_better)
        ## 
        s = s.replace('Correctness', 'Acc')
        s = s.replace('prompt gpt-4o ($K=12$)', 'fixed')
        s = s.replace('$(u_1,\cdots,u_K)$ Source', '$\{u_k\}_k$')
        
        # remove table quotes, since put into minipages
        s = s.replace('\\begin{table}[t]', '')
        s = s.replace('\end{table}', '')
        s = s.replace(f'{model_name}', '{\scriptsize ' + f'{model_name}'  + '}')
        
        st = get_string_between(s, '\\midrule\n', '\\bottomrule')
        st = replace_specific_occurrences(st, model_name, [0,2])
        
        L.append(st)

    ## combine different model into a single table
    start_str = '\\midrule\n'
    end_str = '\n\\bottomrule'
    start = s.find(start_str) + len(start_str)
    end = s.find(end_str, start)
    s = s[:start] + '\n\\midrule\n'.join(L) + s[end:]

    for k in ['survey', 'on-the-fly', 'fixed']:
        s = s.replace(f'{k}', '{\scriptsize ' + f'{k}'  + '}')

    return s


def plt_reliability_diagram_vary_model(
        task_name_list=['truthfulqa'],
        dcp_source='survey',
        calibrate_over_list=['dist', 'dist'],
        answers_subset_list=['test', 'test'],
        model_name_list=['gpt-4o-mini'],
        assets_path='./',
        ylabel=None,
    ):

    nrows = len(answers_subset_list)
    if len(task_name_list) != nrows:
        assert(len(task_name_list) == 1)
        task_name_list = task_name_list * nrows

    cols = ['Model', 'Task', r'$(u_1,\cdots,u_K)$ Source']
    cols += ['Correctness', 'ECE', 'BS']

    exp_dirs = [
        # 'results/gpt-3.5-turbo-0125',
        'results/gpt-4o-mini',
        'results/gpt-4o-2024-08-06',
        'results/claude-3-haiku-20240307',
        'results/claude-3-5-sonnet-20240620',
        'results/gemini-1.5-flash-002',
        'results/gemini-1.5-pro-002',
    ]

    if dcp_source == 'on-the-fly':
        has_dcp_source = lambda x: 'p=verbconfv3_src=unspecified:beta' in x
    elif dcp_source == 'survey':
        has_dcp_source = lambda x: 'p=verbconf_src=social:media:poll' in x
    elif dcp_source.startswith('fixed'):
        match = re.search(r"K=(\d+)", dcp_source)
        if match:
            size = int(match.group(1))
        else:
            size = 12
        match = re.search(r'\(([^)]+)\)$', dcp_source)
        if match:
            dcp_source_model = match.group(1)
        else:
            dcp_source_model = 'gpt-4o'
        print(f'{dcp_source} -> (K: {size}, LLM: {dcp_source_model})')
        has_dcp_source = lambda x: f'p=verbconf_src=prompt:{dcp_source_model}:size={size}' in x
    else:
        raise ValueError(f'Invalid dcp_source: {dcp_source}')

    run_path_filter_fn = lambda x: \
        any([t in x for t in task_name_list]) and \
        has_dcp_source(x)

    info = get_eval_results(
        exp_dirs=exp_dirs,
        cols=cols,
        run_path_filter_fn=run_path_filter_fn,
        group_metrics_by_task=False,
        sort_df=None,
    )
    df = info['df']
    rs = info['taskresults']

    if len(rs) != len(model_name_list) * len(set(task_name_list)):
        rs_repr = '\n'.join([str(r) for r in rs])
        raise ValueError(f"size of results not equal to number of models. rs = \n {rs_repr}")

    rs_for_each_row_list = []
    for i, task_name in enumerate(task_name_list):
        rs = [r for r in info['taskresults'] if r.task_name == task_name]
        rs = sorted(rs, key=lambda x: model_name_list.index(x.df.iloc[0]['Model']))
        print(task_name, len(rs))
        rs_for_each_row_list.append(rs)

    plt_params = {
        'font.size': 25,
        'font.family': 'Times New Roman',
        "legend.frameon": False,
        "axes.titlepad": 15,
        # bar plot
        "patch.force_edgecolor": False,
        'patch.linewidth': 2,  # Edge line width for bar patches
        # tick label size
        'xtick.labelsize': 15,
        'ytick.labelsize': 15,
    }

    with plt.rc_context(plt_params):

        nrows = len(calibrate_over_list)
        ncols = len(model_name_list)
        fig, axs=plt.subplots(nrows, ncols, figsize=(4*ncols, 4*nrows), sharex=True, sharey=True)
        
        for axj in range(ncols):
            for axi in range(nrows):
                calibrate_over = calibrate_over_list[axi]
                answers_subset = answers_subset_list[axi]
                rs = rs_for_each_row_list[axi]
                r = rs[axj]
                rd = r.get_answers_list()
                r = rd[answers_subset]
                ax = axs.reshape(nrows, ncols)[axi, axj]
                if calibrate_over == 'prob':
                    bin_data = r.calibration_bin_data_prob
                elif calibrate_over == 'dist':
                    bin_data = r.calibration_bin_data_dist
                plt_reliability_diagram_subplot(
                    ax, bin_data, **get_reliability_diagram_subplot_kwargs(calibrate_over))
                ax.grid()
                
            ax.set_aspect(aspect='auto', adjustable='box') 
        
        for axj in range(ncols):
            title = model_name_list[axj]
            # fontsize = 40 if len(title) <= 13 else 30
            fontsize = 32
            axs.reshape(nrows, ncols)[0, axj].set_title(title, fontsize=fontsize)

        # set ylabel if use different task_name for each row.
        if len(set(task_name_list)) > 1:
            for axi in range(nrows):
                task_name = task_name_list[axi]
                if task_name == 'sciq': task_name_str = 'SciQ'
                elif task_name == 'truthfulqa': task_name_str = 'TruthfulQA'
                axs.reshape(nrows, ncols)[axi, 0].set_ylabel(task_name_str, fontsize=45)


        # if ylabel is None:
        #     # ylabel = f'{task_name_str}\n{dcp_source}'
        #     ylabel = task_name_str

        if ylabel is not None:
            fig.supylabel(ylabel, x=0, ha='left', va='center', fontsize=50)
            
        fig.tight_layout(pad=.1)


    return fig


def plt_ablate_ot_calibration_hyperparameters(
        save_dir,
        eps_list,
        gamma_list,
        fig_path=None,
    ):
    from evaluate import TaskResult

    options_list = itertools.product(
        eps_list, 
        gamma_list
    )

    r = TaskResult(save_dir)
    rs = r.get_answers_list()

    ot_type_list = []
    for eps, gamma in options_list:
        ot_type = f'fn=unbsinkstab_eps={eps}_gamma={gamma}_div=kl'
        ot_type_list.append(ot_type)

    calibrate_ot_on_task_result(
        r=r,
        ot_type_list=ot_type_list,
        refresh=False,
    )

    # load updated answers.
    rs = r.get_answers_list()

    A = np.zeros((len(eps_list), len(gamma_list)))
    for i, eps in enumerate(eps_list):
        for j, gamma in enumerate(gamma_list):
            k = f'test_fn=unbsinkstab_eps={eps}_gamma={gamma}_div=kl'
            if k not in rs:
                v = np.nan
            else:
                v = rs[k].metrics['ece_dist']
            A[i,j] = v


    plt_params = {
        'font.size': 25,
        'font.family': 'Times New Roman',
        "legend.frameon": False,
        "axes.titlepad": 10,
        # tick label size
        'xtick.labelsize': 15,
        'ytick.labelsize': 15,
    }

    with plt.rc_context(plt_params):

        fig, ax = plt.subplots(1, 1, figsize=(4, 4))
        im = ax.imshow(A, cmap='bwr')
        cbar = fig.colorbar(im, cax=plt_scaled_colobar_ax(ax))
        cbar.locator = MaxNLocator(nbins=4)
        cbar.update_ticks()
        
        for i in range(A.shape[0]):
            for j in range(A.shape[1]):
                ax.annotate(f'{A[i,j]:.2f}' if not np.isnan(A[i,j]) else '', xy=(j, i),
                            fontsize=12,
                            horizontalalignment='center',
                            verticalalignment='center')
        
        ax.set_title('ECE')
        ax.set_yticks(np.arange(len(eps_list)), eps_list)
        ax.set_xticks(np.arange(len(gamma_list)), gamma_list, ha='left', rotation=-30) 

        ax.set_ylabel(r'$\epsilon$', fontsize=25)
        ax.set_xlabel(r'$\tau_2$', fontsize=25)
    
    if fig_path is not None:
        fig.savefig(fig_path, dpi=300, bbox_inches='tight')
        print(f'Saved to fig_path: {fig_path}')

    return fig


def plt_ablate_ot_calibration_hyperparameters_all():
    """Ablate OT calibration epsilon, tau_2 table.  """

    model_name = 'gpt-4o-2024-08-06'
    for task_name in ['truthfulqa', 'sciq']:
        save_dir = f'results/{model_name}/{task_name}_p=verbconf_src=prompt:gpt-4o:size=12'
        eps_list = [1e-4, 1e-3, 1e-2, 3e-2, 1e-1, 3e-1, 1]
        gamma_list = [1e-2, 1e-1, 3e-1, 1, 3, 10, 100]
        fig_path = os.path.join(
            assets_path, 
            f'ablation_ot_calibration_hyperparams_task={task_name}_m={model_name}_p=verbconf_src=prompt:gpt-4o:size=12.png')

        plt_ablate_ot_calibration_hyperparameters(
            save_dir=save_dir,
            eps_list=eps_list,
            gamma_list=gamma_list,
            fig_path=fig_path,
        )


def prepare_llm_ot_calibration_table():
    """Table that contains
        - verbconf
        - verbconf with OT calibrate
        - verbprobtop1
        - verbprobtop1 calibrated with platt scaling
    """
    from calibration import plt_ablate_ot_calibration_hyperparameters, calibrate_probs_on_verbprob, calibrate_probs_on_verbconf

    task_name_list = ['sciq', 'truthfulqa']

    ## OT calibration
    model_name_list = [
        "gpt-4o-mini",
        "gpt-4o-2024-08-06",
        'claude-3-haiku-20240307',
        'claude-3-5-sonnet-20240620',
        'gemini-1.5-flash-002',
        'gemini-1.5-pro-002',
    ]

    for model_name in model_name_list:
        for task_name in task_name_list:
            save_dir = f'results/{model_name}/{task_name}_p=verbconf_src=prompt:gpt-4o:size=12'
            plt_ablate_ot_calibration_hyperparameters(
                save_dir=save_dir,
                eps_list=[1e-3],
                gamma_list=[1],
            )


    ## platt scaling on verbprob
    calibration_type = 'platt'
    model_name_list = [
        "gpt-4o-2024-08-06",
        'claude-3-5-sonnet-20240620',
        'gemini-1.5-pro-002',
    ]

    for model_name in model_name_list:
        for task_name in task_name_list:
            save_dir = f'results/{model_name}/{task_name}_p=verbprobtop1_src=unspecified:delta'
            r = TaskResult(save_dir)
            calibrate_probs_on_verbprob(
                r,
                calibration_type=calibration_type,
                refresh=True,
            )

    ## platt scaling / histogram binning on verbconf
    model_name_list = [
        "gpt-4o-2024-08-06",
        'claude-3-5-sonnet-20240620',
        'gemini-1.5-pro-002',
    ]
    calibration_type_list = ['platt', 'histbinning']

    for model_name in model_name_list:
        for task_name in task_name_list:
            save_dir = f'results/{model_name}/{task_name}_p=verbconf_src=prompt:gpt-4o:size=12'
            r = TaskResult(save_dir)
            for calibration_type in calibration_type_list:
                calibrate_probs_on_verbconf(
                    r,
                    calibration_type=calibration_type,
                    refresh=True,
                )


    ## convert OT calibration result to verbprobtop1-like data subset
    # so that the metrics {ECE,BS} are computed in the same mannter.
    model_name_list = [
        "gpt-4o-2024-08-06",
        'claude-3-5-sonnet-20240620',
        'gemini-1.5-pro-002',
    ]
    data_subset = 'test_fn=unbsinkstab_eps=0.001_gamma=1_div=kl'

    for model_name in model_name_list:
        for task_name in task_name_list:
            save_dir = f'results/{model_name}/{task_name}_p=verbconf_src=prompt:gpt-4o:size=12'
            r = TaskResult(save_dir)
            r_test = r.get_answers_list()[data_subset]
            probs = r_test.classification_data['u_prob'][(r_test.model_name, 'pred')].round(3)
            probs = [str(x) for x in probs]
            # modify answers with transported answers
            answers = []
            for d, confidence in zip(r_test.answers, probs):
                d = d.copy()
                d['confidence'] = confidence
                answers.append(d)
            r.save_answers(
                data_subset + '_asverbprob',
                answers,
                force=True,
                additional_config={'prompt_type': 'verbprobtop1', 'certainty_phrase_source': 'unspecified_delta'},
            )


def table_llm_calibration():
    """Main table that compares performance of LLM verbalized confidence with baselines"""


    model_name_list = [
        'gpt-4o-2024-08-06',
        'claude-3-5-sonnet-20240620',
        'gemini-1.5-pro-002',
    ]
    task_name_list = ['truthfulqa', 'sciq']
    cols = ['Task', 'Model', 'Prompt', 'calibrate_type'] + ['Correctness', 'ece_prob', 'BS']
    dfs = []


    ## verbconf with [scaling, binning, ot]
    run_paths = []
    data_subsets_list = [
        'test', 
        'test_platt',
        'test_histbinning',
        # 'test_fn=unbsinkstab_eps=0.001_gamma=1_div=kl',
        # ## asverbprob: convert confidence -> mean of dist for comparison.
        'test_fn=unbsinkstab_eps=0.001_gamma=1_div=kl_asverbprob',
    ]
    for model_name, task_name in itertools.product(model_name_list, task_name_list):
        save_dir = f'results/{model_name}/{task_name}_p=verbconf_src=prompt:gpt-4o:size=12'
        for data_subsets in data_subsets_list:
            run_paths.append(os.path.join(save_dir, data_subsets))

    info = get_eval_results(
        exp_dirs=[],
        run_paths=run_paths,
        cols=cols,
        group_metrics_by_task=True,
        sort_df=None,
    )
    df = info['df']
    df[('Prompt', '')] = 'Verb. Conf.' # 'Verb. Prob.' -> 'Verb. Conf.' since scaling/binning changed it.
    dfs.append(df)

    ## verbprobtop1 w. scaling.
    run_paths = []
    data_subsets_list = ['test', 'test_platt']
    for model_name, task_name in itertools.product(model_name_list, task_name_list):
        save_dir = f'results/{model_name}/{task_name}_p=verbprobtop1_src=unspecified:delta'
        for data_subsets in data_subsets_list:
            run_paths.append(os.path.join(save_dir, data_subsets))

    info = get_eval_results(
        exp_dirs=[],
        run_paths=run_paths,
        cols=cols,
        group_metrics_by_task=True,
        sort_df=None,
    )
    df = info['df']
    dfs.append(df)

    df = pd.concat(dfs)
    df = df.reset_index(drop=True)


    # Function to map the values
    def is_calibrated_fn(value):
        if value == "":
            return r"\textcolor{red}{$\times$}"  # Red cross symbol
        else:
            if value == 'Platt': value = r"{\footnotesize scaling}"
            if value == 'Hist Binning': value = r"{\footnotesize binning}"
            if value == 'OT': value = r"{\footnotesize \textbf{ours}}"
            return r"\textcolor{green}{$\checkmark$} (" + value + ")"

    model_name_list = [
        'gpt-4o',
        'claude-3.5-sonnet',
        'gemini-1.5-pro',
    ]

    L = []
    for model_name in model_name_list:
        
        dft = df[df[('Model','')]==model_name].copy()
        # sort 
        ordering = []
        ordering += [('Verb. Prob.', ''),('Verb. Prob.', 'Platt')]
        ordering += [('Verb. Conf.', x) for x in ['', 'Platt', 'Hist Binning', 'OT']]
        dft['ord'] = pd.Categorical(list(zip(dft['Prompt'], dft['calibrate_type'])), categories=ordering, ordered=True)
        dft = dft.sort_values('ord').drop(columns=['ord'])
        dft.reset_index(drop=True)
        dft[('calibrate_type', '')] = dft[('calibrate_type', '')].apply(is_calibrated_fn)


        # ### bold cells
        # formatter = get_default_bold_formatter(
        #     dft,
        #     cols_take_max=['Acc', 'Correctness'],
        #     cols_take_min=['ECE', 'BS'],
        # )
        # s = df_to_latex(dft, 
        #                 column_format=None,
        #                 formatter=formatter,
        #             )
        ### color cells 
        from matplotlib.colors import to_rgb, LinearSegmentedColormap
        def create_custom_blue_cmap():
            return LinearSegmentedColormap.from_list("custom_blue", ['#FFFFFF', '#E6F3FF', '#99D6FF', '#4DB8FF'])
        cmap = create_custom_blue_cmap()
        higher_better = [x for x in list(dft.columns) if x[1]]
        higher_better = dict({x: (True if x[1] in ['Correctness'] else False) for x in higher_better })
        subset_to_color = [x for x in df.columns if x[1] in ['BS', 'ece_prob', 'ECE']]
        subset_to_color = None
        s = df_to_latex_with_colors(dft.round(2), cmap=cmap, alpha=.75, low=0, high=.7,
                                higher_better=higher_better, subset=None)
        ### 
        
        s = s.replace('calibrate_type', 'Calibrated?')
        s = s.replace('Correctness', 'Acc')
        s = s.replace('Verb. Conf.', r'{\footnotesize Phrase}')
        s = s.replace('Verb. Prob.', r'{\footnotesize Probability}')
        s = s.replace('Prompt', 'Verbalize')
        s = s.replace(r'{\footnotesize Probability}', r'{\scriptsize Probability}')
        s = s.replace(r'{\footnotesize Phrase}', r'{\scriptsize Phrase}')
        s = s.replace('0000 &', ' &')
        st = get_string_between(s, '\\midrule\n', '\\bottomrule')
        
        # st = replace_specific_occurrences(st, model_name, [1,2,3])
        # st = replace_specific_occurrences(st, 'Probability', [1])
        # st = replace_specific_occurrences(st, 'Certainty Phrase', [1])
        st = replace_specific_occurrences(st, model_name, [1,2,3,4,5])
        st = replace_specific_occurrences(st, 'Probability', [1])
        st = replace_specific_occurrences(st, 'Phrase', [1,2,3])
        st = st.replace(model_name, '{\\footnotesize ' + model_name + '}')

        L.append(st)


    ## combine different model into a single table
    start_str = '\\midrule\n'
    end_str = '\n\\bottomrule'
    start = s.find(start_str) + len(start_str)
    end = s.find(end_str, start)
    s = s[:start] + '\n\\midrule\n'.join(L) + s[end:]

    s = s.replace('ece_prob', 'ECE')

    for k in ['ECE ', 'BS ']:
        s = s.replace(k, k + r'\scalebox{0.7}{$\downarrow$} ')
    for k in ['Acc ']:
        s = s.replace(k, k + r'\scalebox{0.7}{$\uparrow$} ')
        

    print(s)


def tab_show_probability_density_function_of_confidence_distributions(assets_path):
    from calibration import get_dcp_dist, plt_dcp_distribution_density
    ncols = 8

    ## radiology
    source = 'radiology'
    for dcp_dist_type in ['beta_fit_mm', 'beta_fit_mle', 'mixture']:
        print(dcp_dist_type)
        dcp_to_dist = get_dcp_dist(source='radiology', distribution_type=dcp_dist_type)
        dcp_to_dist = {k:v for k,v in dcp_to_dist.items() if k not in ['absent', 'present', 'diagnostic of', 'other']}
        
        fig = plt_dcp_distribution_density(dcp_to_dist, ncols=ncols)
        fig.savefig(os.path.join(assets_path, f'density_function_dcp_{source}_{dcp_dist_type}.png'), dpi=300, bbox_inches='tight')

    ## social media
    source = 'social_media_poll'
    distribution_type='beta_fit_mm'
    dcp_to_dist = get_dcp_dist(source, distribution_type=distribution_type)
    dcp_to_dist = {k:v for k,v in dcp_to_dist.items() if k not in ['Better than Even']}
    fig = plt_dcp_distribution_density(dcp_to_dist, ncols=ncols)
    fig.savefig(os.path.join(assets_path, f'density_function_dcp_social_media_{distribution_type}.png'), dpi=300, bbox_inches='tight')

    ## gpt-4o generated
    source_list = [f'prompt_gpt-4o_size={K}' for K in [4,8,12,16]]

    for source in source_list:
        dcp_to_dist = get_dcp_dist(source)
        fig = plt_dcp_distribution_density(dcp_to_dist, ncols=ncols)
        fig.savefig(os.path.join(assets_path, f'density_function_dcp_source={source}.png'), dpi=300, bbox_inches='tight')